from .. import *
from .base import *

class UProgressBar(UBase):
    def __init__(self, progress, done=1):
        super().__init__()
        self.progress = progress * 100
        self.done = done * 100

    def update(self, prev):
        if prev and hasattr(prev, "ui"):
            self.ui = prev.ui
            self.ui.set_completion(self.progress)
        else:
            self.ui = urwid.ProgressBar('pb:normal', 'pb:complete', self.progress, self.done)



from .. import *
from .base import *

class FRow(FBase):
    def update(self, prev):
        if prev and hasattr(prev, "ui"):
            self.ui = prev.ui
        else:
            self.ui = ft.Row(expand=self.layout_weight)

    def addChild(self, idx, child):
        self.ui.controls.append(child.ui)
        try:
            self.ui.update()
        except:
            pass

    def removeChild(self, idx, child):
        self.ui.controls.remove(child.ui)
        self.ui.update()

class FColumn(FBase):
    def update(self, prev):
        if prev and hasattr(prev, "ui"):
            self.ui = prev.ui
        else:
            self.ui = ft.Column(expand=self.layout_weight)

    def addChild(self, idx, child):
        self.ui.controls.append(child.ui)
        try:
            self.ui.update()
        except:
            pass

    def removeChild(self, idx, child):
        self.ui.controls.remove(child.ui)
        self.ui.update()



from .. import *
import tkinter as tk
from tkinter import ttk
class TkBaseWidget(PUINode):
    def __init__(self, layout=None, side=None, **kwargs):
        super().__init__()
        self.layout_type = layout
        self.side = side
        self.kwargs = kwargs

    @property
    def tkparent(self):
        parent = self.parent
        while not isinstance(parent, TkBaseWidget):
            parent = parent.parent
        return parent

    def destroy(self):
        if self.ui:
            self.ui.destroy()
            self.ui = None


from .. import *
from .base import *

class UColumns(UBase):
    def update(self, prev):
        if prev and hasattr(prev, "ui"):
            self.ui = prev.ui
        else:
            self.ui = urwid.Columns([])

    def addChild(self, idx, child):
        self.ui.contents.append((child.ui, self.ui.options()))

    def removeChild(self, idx, child):
        self.ui.contents.pop(idx)


class UPile(UBase):
    def update(self, prev):
        if prev and hasattr(prev, "ui"):
            self.ui = prev.ui
        else:
            self.ui = urwid.Pile([])

    def addChild(self, idx, child):
        self.ui.contents.append((child.ui, self.ui.options()))

    def removeChild(self, idx, child):
        self.ui.contents.pop(idx)



# dprint = print
dprint = lambda *x: x

def recur_delete(node, idx, child):
    for sidx,sc in enumerate(child.children):
        recur_delete(child, sidx, sc)
    node.removeChild(idx, child)
    child.destroy()

def sync(node, oldDOM, newDOM, children_first):
    dprint("syncing", node.key, len(oldDOM), len(newDOM))
    dprint("  ===OLD===")
    for c in oldDOM:
        dprint("   ", c.key)
    dprint("  ===NEW===")
    for c in newDOM:
        dprint("   ", c.key)
    oldMap = [x.key for x in oldDOM]

    skipHead = 0
    for i in range(0, min(len(oldDOM), len(newDOM))):
        if oldDOM[i].key == newDOM[i].key:
            oldMap[i] = None
            old = oldDOM[i]
            new = newDOM[i]
            try:
                new.update(old)
            except:
                import traceback
                print("## <ERROR OF sync() >")
                print(node.key)
                traceback.print_exc()
                print("## </ERROR OF sync()>")

            if not new.terminal:
                sync(new, old.children, new.children, children_first)
            skipHead += 1
        else:
            break

    skipTail = 0
    # for i in range(0, min(len(oldDOM)-skipHead, len(newDOM)-skipHead)):
    #     if oldDOM[-1-i].key == newDOM[-1-i].key:
    #         oldMap[-1-i] = None
    #         old = oldDOM[-1-i]
    #         new = newDOM[-1-i]
    #         new.update(old)
    #         sync(new, old.children, new.children)
    #         skipTail += 1
    #     else:
    #         break

    for i, new in enumerate(newDOM[skipHead:len(newDOM)-skipTail]):
        new_idx = skipHead + i
        try:
            old_idx = oldMap.index(new.key)
        except:
            old_idx = -1
        if old_idx >= 0:
            oldMap[old_idx] = None
            old = oldDOM[old_idx]
            node.removeChild(old_idx, old)
            new.update(old)
            if not new.terminal and children_first:
                sync(new, old.children, new.children, children_first)
            node.addChild(new_idx, old)
            if not new.terminal and not children_first:
                sync(new, old.children, new.children, children_first)
        else:
            new.update(None)
            if not new.terminal and children_first:
                sync(new, [], new.children, children_first)
            node.addChild(new_idx, new)
            if not new.terminal and not children_first:
                sync(new, [], new.children, children_first)

    for old_idx, key in enumerate(oldMap):
        if key:
            old = oldDOM[old_idx]
            recur_delete(node, old_idx, old)



from .. import *
from .base import *

class TkApplication(PUIView):
    def redraw(self):
        if self.ui:
            self.ui.after(0, self.update)
        else:
            self.update()

    def update(self, prev=None):
        if prev and hasattr(prev, "ui"):
            self.ui = prev.ui
        else:
            self.ui = tk.Tk()
            self.ui.withdraw()

        super().update(prev)

    def addChild(self, idx, child):
        pass

    def removeChild(self, idx, child):
        child.ui.destroy()

    def start(self):
        self.ui.mainloop()



from .. import *
from .base import *

class UText(UBase):
    def __init__(self, text, **kwargs):
        super().__init__(**kwargs)
        self.text = text

    def update(self, prev):
        if prev and hasattr(prev, "ui"):
            self.ui = prev.ui
            self.ui.set_text(self.text)
        else:
            self.ui = urwid.Text(self.text)



from .. import *
from .base import *
class TkCanvas(TkBaseWidget):
    terminal = True
    def __init__(self, size=None, **kwargs):
        super().__init__(**kwargs)
        self.size = size

    def update(self, prev):
        if prev and hasattr(prev, "ui"):
            self.ui = prev.ui
        else:
            self.ui = tk.Canvas(self.parent.ui, **self.kwargs)
        self.ui.delete("all")

        for c in self.children:
            c.update(None)


class TkCanvasText(PUINode):
    def __init__(self, x, y, text):
        super().__init__()
        self.x = x
        self.y = y
        self.text = text

    def update(self, prev):
        self.parent.ui.create_text(self.x, self.y, text=self.text)

class TkCanvasLine(PUINode):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.args = args
        self.kwargs = kwargs

    def update(self, prev):
        self.parent.ui.create_line(*self.args, **self.kwargs)


from .. import *
from .base import *

class TkWindow(TkBaseWidget):
    def __init__(self, title=None, size=None, resizable=None):
        super().__init__()
        self.title = title
        self.size = size
        self.resizable = resizable

    def update(self, prev):
        if prev and hasattr(prev, "ui"):
            self.ui = prev.ui
        else:
            self.ui = tk.Toplevel(self.parent.ui)

        if not self.title is None:
            self.ui.title(self.title)
        if not self.size is None:
            self.ui.geometry("x".join([str(v) for v in self.size]))

        if type(self.resizable) is bool:
            resizable = (self.resizable, self.resizable)
        elif self.resizable:
            resizable = self.resizable
        else:
            resizable = (True, True)
        self.ui.resizable(*resizable)
        self.ui.iconbitmap('icon.ico')

    def addChild(self, idx, child):
        if child.layout=="pack":
            child.ui.pack(side=child.side, **child.kwargs)
        elif child.layout=="grid":
            child.ui.grid(**child.kwargs)
        elif child.layout=="place":
            child.ui.place(*child.kwargs)
        else:
            child.ui.pack(fill=tk.BOTH, expand=True)

    def removeChild(self, idx, child):
        if child.layout=="pack":
            child.ui.pack_forget()
        elif child.layout=="grid":
            child.ui.grid_forget()
        elif child.layout=="place":
            child.ui.place_forget()
        else:
            child.ui.pack_forget()



from .. import *
from .base import *

class QtHBox(QtBaseLayout):
    def update(self, prev):
        if prev and hasattr(prev, "ui"):
            self.ui = prev.ui
        else:
            from PyQt5 import QtWidgets
            self.ui = QtWidgets.QHBoxLayout()
            self.ui.setObjectName(self.key)

class QtVBox(QtBaseLayout):
    def update(self, prev):
        if prev and hasattr(prev, "ui"):
            self.ui = prev.ui
        else:
            self.ui = QtWidgets.QVBoxLayout()
            self.ui.setObjectName(self.key)



from .. import *
from .base import *

class FProgressBar(FBase):
    def __init__(self, progress):
        super().__init__()
        self.progress = progress

    def update(self, prev):
        if prev and hasattr(prev, "ui"):
            self.ui = prev.ui
            self.ui.value = self.progress
            self.ui.update()
        else:
            self.ui = ft.ProgressBar()
            self.ui.value = self.progress



from .. import *
from .base import *

class FElevatedButton(FBase):
    def __init__(self, text, callback=None, **kwargs):
        super().__init__(**kwargs)
        self.text = text
        self.callback = callback
        self.kwargs = kwargs

    def update(self, prev):
        if prev and hasattr(prev, "ui"):
            self.ui = prev.ui
            self.ui.text = self.text
            self.ui.on_click = self.on_click
            self.ui.update()
        else:
            self.ui = ft.ElevatedButton(text=self.text, on_click=self.on_click, expand=self.layout_weight, **self.kwargs)

    def on_click(self, *args):
        self.callback()


from .. import *
from .base import *

class UButton(UBase):
    def __init__(self, text, callback=None):
        super().__init__()
        self.text = text
        self.callback = callback

    def update(self, prev):
        if prev and hasattr(prev, "ui"):
            self.ui = prev.ui
            self.btn = prev.btn
            self.btn.set_label(self.text)
            self.btn.pui_cb = self.callback
        else:
            self.btn = urwid.Button(self.text, on_press=self.on_press)
            self.btn.pui_cb = self.callback
            self.ui = urwid.AttrWrap(self.btn, 'btn', 'btn:focus')

    def on_press(self, ui):
        if ui.pui_cb:
            ui.pui_cb()


from .. import *
from .base import *
class TkLabel(TkBaseWidget):
    def __init__(self, text, **kwargs):
        super().__init__(**kwargs)
        self.text = text

    def update(self, prev):
        if prev and hasattr(prev, "ui"):
            self.ui = prev.ui
            self.ui.config(text = self.text)
        else:
            self.ui = tk.Label(self.parent.ui, text=self.text, **self.kwargs)



from .application import *
from .button import *
from .canvas import *
from .label import *
from .layout import *
from .progressbar import *
from .textfield import *
from .window import *

Application = QtApplication
Window = QtWindow
HBox = QtHBox
VBox = QtVBox
Label = QtLabel
Button = QtButton
Canvas = QtCanvas
CanvasText = QtCanvasText
CanvasLine = QtCanvasLine
TextField = QtLineEdit
ProgressBar = QtProgressBar


from .. import *
from .base import *

class FWindow(PUINode):
    def __init__(self, title=None, size=None):
        super().__init__()
        self.children_first = True
        self.title = title
        self.size = size

    def update(self, prev=None):
        self.parent.ui.title = self.title
        super().update(prev)

    def addChild(self, idx, child):
        self.parent.ui.add(child.ui)

    def removeChild(self, idx, child):
        self.parent.ui.remove(child.ui)



from .. import *
from .base import *

from PyQt5 import QtWidgets
from PyQt5.QtGui import QPainter, QColor
from PyQt5.QtCore import QPoint

class PUIQtCanvas(QtWidgets.QWidget):
    def __init__(self, puinode):
        self.puinode = puinode
        super().__init__()

    def paintEvent(self, event):
        qpainter = QPainter()
        qpainter.begin(self)

        for c in self.puinode.children:
            c.draw(qpainter)

        qpainter.end()

class QtCanvas(QtBaseWidget):
    def __init__(self, size=None, **kwargs):
        super().__init__(**kwargs)
        self.size = size

    def update(self, prev):
        if prev and hasattr(prev, "ui"):
            self.ui = prev.ui
            self.ui.puinode = self
        else:
            self.ui = PUIQtCanvas(self)
        x = 0
        y = 0
        w = 0
        h = 0
        if not self.size is None:
            w, h = self.size
        self.ui.setGeometry(x, y, w, h)
        self.ui.update()

class QtCanvasText(PUINode):
    def __init__(self, x, y, text):
        super().__init__()
        self.x = x
        self.y = y
        self.text = text

    def update(self, prev):
        pass

    def draw(self, qpainter):
        qpainter.drawText(QPoint(int(self.x), int(self.y)), self.text)

class QtCanvasLine(PUINode):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.args = args
        self.kwargs = kwargs

    def update(self, prev):
        pass

    def draw(self, qpainter):
        qpainter.drawLine(*self.args)


from .. import *
from .base import *

class QtLineEdit(QtBaseWidget):
    def __init__(self, model):
        super().__init__()
        self.model = model

    def update(self, prev):
        value = self.model.value
        if prev and hasattr(prev, "ui"):
            self.ui = prev.ui
            if prev.last_value != value:
                self.ui.textChanged.disconnect()
                self.ui.setText(str(value))
                self.ui.textChanged.connect(self.on_textchanged)
            self.last_value = value
        else:
            self.last_value = value
            self.ui = QtWidgets.QLineEdit()
            self.ui.setText(str(value))
            self.ui.textChanged.connect(self.on_textchanged)

    def on_textchanged(self):
        self.model.value = self.ui.text()



from .. import *
from .base import *

class QtWindow(PUINode):
    def __init__(self, title=None, size=None):
        super().__init__()
        self.title = title
        self.size = size

    def update(self, prev=None):
        if prev and hasattr(prev, "ui"):
            self.ui = prev.ui
        else:
            from PySide6 import QtWidgets
            self.ui = QtWidgets.QWidget()
            self.ui.setObjectName("Window")
            self.box = QtWidgets.QBoxLayout(QtWidgets.QBoxLayout.Direction.LeftToRight)
            self.ui.setLayout(self.box)

        if not self.title is None:
            self.ui.setWindowTitle(self.title)
        if not self.size is None:
            self.ui.resize(*self.size)

        super().update(prev)

    def addChild(self, idx, child):
        if isinstance(child, QtBaseLayout):
            self.box.addLayout(child.ui)
        elif isinstance(child, QtBaseWidget):
            self.box.addWidget(child.ui)
        else:
            self.addChild(idx, child.children[0])

    def removeChild(self, idx, child):
        if isinstance(child, QtBaseLayout):
            self.box.removeItem(child.ui)
        elif isinstance(child, QtBaseWidget):
            child.ui.setParent(None)
        else:
            self.removeChild(idx, child.children[0])



from .application import *
from .button import *
from .label import *
from .layout import *
from .progressbar import *
from .textfield import *
from .window import *
from .. import NotImplementedNode

Application = FApplication
Window = FWindow
HBox = FRow
VBox = FColumn
Label = FText
Button = FElevatedButton
Canvas = NotImplementedNode
CanvasText = NotImplementedNode
CanvasLine = NotImplementedNode
TextField = FTextField
ProgressBar = FProgressBar


from .. import *
import urwid
import sys

class UWindow(PUINode):
    def __init__(self, title=None, size=None):
        super().__init__()
        self.title = title
        self.size = size

    @property
    def ui(self):
        return self.children[0].ui

    @ui.setter
    def ui(self, new_ui):
        pass



from .. import *
from .base import *

class TkProgressBar(TkBaseWidget):
    def __init__(self, progress, maximum=1):
        super().__init__()
        self.progress = progress*100
        self.maximum = maximum*100

    def update(self, prev):
        if prev and hasattr(prev, "ui"):
            self.ui = prev.ui
        else:
            self.ui = ttk.Progressbar(self.parent.ui)
        self.ui["maximum"] = self.maximum
        self.ui["value"] = self.progress



from .. import *
from .base import *

class QtLabel(QtBaseWidget):
    def __init__(self, text):
        super().__init__()
        self.text = text

    def update(self, prev):
        if prev and hasattr(prev, "ui"):
            self.ui = prev.ui
            self.ui.setText(self.text)
        else:
            self.ui = QtWidgets.QLabel(self.text)



from .. import *

from PySide6 import QtCore, QtWidgets

class QtViewSignal(QtCore.QObject):
    redraw = QtCore.Signal()

class QPUIView(PUIView):
    def __init__(self):
        super().__init__()
        self.signal = QtViewSignal()
        self.signal.redraw.connect(self.update)

    def redraw(self):
        self.signal.redraw.emit()

class QtBaseWidget(PUINode):
    terminal = True
    def destroy(self):
        self.ui.deleteLater()

class QtBaseLayout(PUINode):
    def addChild(self, idx, child):
        if isinstance(child, QtBaseLayout):
            self.ui.addLayout(child.ui)
        elif isinstance(child, QtBaseWidget):
            params = {}
            if not child.layout_weight is None:
                params["stretch"] = child.layout_weight
            self.ui.addWidget(child.ui, **params)
        elif child.children:
            self.addChild(idx, child.children[0])

    def removeChild(self, idx, child):
        if isinstance(child, QtBaseLayout):
            self.box.removeItem(child.ui)
        elif isinstance(child, QtBaseWidget):
            child.ui.setParent(None)
        elif child.children:
            self.removeChild(idx, child.children[0])



from .. import *
from .base import *

from PySide6 import QtWidgets
from PySide6.QtGui import QPainter, QColor
from PySide6.QtCore import QPoint

class PUIQtCanvas(QtWidgets.QWidget):
    def __init__(self, puinode, width=None, height=None):
        self.puinode = puinode
        self.width = width
        self.height = height
        super().__init__()

    def sizeHint(self):
        return QtCore.QSize(self.width, self.height)

    def paintEvent(self, event):
        qpainter = QPainter()
        qpainter.begin(self)

        for c in self.puinode.children:
            c.draw(qpainter)

        qpainter.end()

class QtCanvas(QtBaseWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ui = None

    def update(self, prev):
        if prev and hasattr(prev, "ui"):
            self.ui = prev.ui
            self.ui.puinode = self
        else:
            self.ui = PUIQtCanvas(self, self.layout_width or 0, self.layout_height or 0)
        self.ui.resize(self.layout_width or 0, self.layout_height or 0)
        self.ui.update()

class QtCanvasText(PUINode):
    def __init__(self, x, y, text):
        super().__init__()
        self.x = x
        self.y = y
        self.text = text

    def update(self, prev):
        pass

    def draw(self, qpainter):
        qpainter.drawText(QPoint(int(self.x), int(self.y)), self.text)

class QtCanvasLine(PUINode):
    def __init__(self, x1, y1, x2, y2, color=None, width=None):
        super().__init__()
        self.x1 = x1
        self.y1 = y1
        self.x2 = x2
        self.y2 = y2
        self.color = color
        self.width = width

    def update(self, prev):
        pass

    def draw(self, qpainter):
        pen = qpainter.pen()
        color = pen.color()
        width = pen.width()
        if not self.color is None:
            pen.setColor(QColor(self.color))
        if not self.width is None:
            pen.setWidth(self.width)
        qpainter.setPen(pen)
        qpainter.drawLine(self.x1, self.y1, self.x2, self.y2)
        pen.setColor(color)
        pen.setWidth(width)
        qpainter.setPen(pen)



from .application import *
from .button import *
from .label import *
from .layout import *
from .progressbar import *
from .window import *
from .. import NotImplementedNode

Application = UApplication
Window = UWindow
HBox = UColumns
VBox = UPile
Label = UText
Button = UButton
Canvas = NotImplementedNode
CanvasText = NotImplementedNode
CanvasLine = NotImplementedNode
TextField = NotImplementedNode
ProgressBar = UProgressBar


from threading import Timer
from .view import *

class TimelineView(PUINode):
    def __init__(self, ttl_sec):
        super().__init__()
        self.ttl_sec = ttl_sec

    def update(self, prev):
        if prev and hasattr(prev, "timer"):
            self.timer = prev.timer
        else:
            self.timer = Timer(self.ttl_sec, self.timer_cb)
            self.timer.start()

    def timer_cb(self):
        try:
            self.root.redraw()
            self.timer = Timer(self.ttl_sec, self.timer_cb)
            self.timer.start()
        except:
            pass

    @property
    def ui(self):
        return self.children[0].ui

    @ui.setter
    def ui(self, new_ui):
        pass

    def destroy(self):
        timer = self.timer
        self.timer = None
        if timer:
            timer.cancel()

    def addChild(self, idx, child):
        self.parent.addChild(idx, child)

    def removeChild(self, idx, child):
        self.parent.removeChild(idx, child)



from .application import *
from .button import *
from .canvas import *
from .label import *
from .layout import *
from .progressbar import *
from .textfield import *
from .window import *

Application = QtApplication
Window = QtWindow
HBox = QtHBox
VBox = QtVBox
Label = QtLabel
Button = QtButton
Canvas = QtCanvas
CanvasText = QtCanvasText
CanvasLine = QtCanvasLine
TextField = QtLineEdit
ProgressBar = QtProgressBar

def PUI(func):
    def func_wrapper(*args):
        class PUIViewWrapper(QPUIView):
            def __init__(self, name):
                self.name = name
                super().__init__()

            def content(self):
                return self.__wrapped_content__()

            def __wrapped_content__(self):
                return func(*args)
        ret = PUIViewWrapper(func.__name__)
        return ret

    return func_wrapper


from .application import *
from .button import *
from .canvas import *
from .label import *
from .layout import *
from .progressbar import *
from .textfield import *
from .window import *

Application = TkApplication
Window = TkWindow
HBox = TkHBox
VBox = TkVBox
Label = TkLabel
Button = TkButton
Canvas = TkCanvas
CanvasText = TkCanvasText
CanvasLine = TkCanvasLine
TextField = TkEntry
ProgressBar = TkProgressBar


from .. import *
from .base import *

class TkHBox(TkBaseWidget):
    def update(self, prev):
        if prev and hasattr(prev, "ui"):
            self.ui = prev.ui
        else:
            self.ui = tk.Frame(self.tkparent.ui)
            self.ui.rowconfigure(0, weight=1)

    def addChild(self, idx, child):
        if isinstance(child, TkBaseWidget):
            child.ui.grid(row=0, column=idx, sticky='nsew')
            if not child.layout_weight is None:
                self.ui.columnconfigure(idx, weight=child.layout_weight)
        elif child.children:
            self.addChild(idx, child.children[0])

    def removeChild(self, idx, child):
        if isinstance(child, TkBaseWidget):
            child.ui.grid_forget()
        elif child.children:
            self.removeChild(idx, child.children[0])

class TkVBox(TkBaseWidget):
    def update(self, prev):
        if prev and hasattr(prev, "ui"):
            self.ui = prev.ui
        else:
            self.ui = tk.Frame(self.tkparent.ui)
            self.ui.config(bg="white")
            self.ui.columnconfigure(0, weight=1)

    def addChild(self, idx, child):
        if isinstance(child, TkBaseWidget):
            child.ui.grid(row=idx, column=0, sticky='nsew')
            if not child.layout_weight is None:
                self.ui.rowconfigure(idx, weight=child.layout_weight)
        elif child.children:
            self.addChild(idx, child.children[0])

    def removeChild(self, idx, child):
        if isinstance(child, TkBaseWidget):
            child.ui.grid_forget()
        elif child.children:
            self.removeChild(idx, child.children[0])





from .. import *
from .base import *
import asyncio

class UApplication(PUIView):
    def __init__(self):
        super().__init__()
        self._palette = [
            # http://urwid.org/reference/display_modules.html#urwid.BaseScreen.register_palette
            # name,          fg,    bg,         mono, fg_high, bg_high
            ('btn',         'white','dark blue'),
            ('btn:focus',   'white','dark red','bold'),
            ('pb:normal',   '',     'dark gray'),
            ('pb:complete', '',     'white'),
        ]
        self.ui = urwid.Filler( urwid.Text(""))
        self.loop = asyncio.get_event_loop()

    def redraw(self):
        self.loop.call_soon(self.update)

    def addChild(self, idx, child):
        if idx>0:
            raise RuntimeError("Urwid only support single window")
        self.ui.set_body(child.ui)

    def on_unhandled_input(self, key):
        pass

    def palette(self, p):
        self._palette = p

    def start(self):
        self.urwid_loop = urwid.MainLoop(
            urwid.Frame(self.ui),
            event_loop=urwid.AsyncioEventLoop(loop=self.loop),
            palette=self._palette,
            unhandled_input=self.on_unhandled_input
        )
        self.urwid_loop.run()
